@@ -10,7 +10,7 @@ use Exporter;
 use SelfLoader;
 use vars qw { $VERSION @ISA %EXPORT_TAGS };
 
-$VERSION = '1.76';
+$VERSION = '1.81';
 @ISA		= qw ( Exporter );
 		     
 %EXPORT_TAGS	= ( ALL => [ qw(
@@ -35,7 +35,7 @@ Exporter::export_ok_tags('ALL');
 sub _match_bracketed($$$$$$);
 sub _match_variable($$);
 sub _match_codeblock($$$$$$$);
-sub _match_quotelike($$$);
+sub _match_quotelike($$$$);
 
 # HANDLE RETURN VALUES IN VARIOUS CONTEXTS
 
@@ -213,7 +213,7 @@ sub _match_bracketed($$$$$$)	# $textref, $pre, $ldel, $qdel, $quotelike, $rdel
 			pos $$textref = $startpos;
 			return;
 		}
-		elsif ($quotelike && _match_quotelike($textref,"",0))
+		elsif ($quotelike && _match_quotelike($textref,"",1,0))
 		{
 			next;
 		}
@@ -246,7 +246,7 @@ sub revbracket($)
 	return $brack;
 }
 
-my $XMLNAME = q{[a-zA-Z_:][a-zA-Z_:.-]*};
+my $XMLNAME = q{[a-zA-Z_:][a-zA-Z0-9_:.-]*};
 
 sub extract_tagged (;$$$$$) # ($text, $opentag, $closetag, $pre, \%options)
 {
@@ -279,6 +279,7 @@ sub extract_tagged (;$$$$$) # ($text, $opentag, $closetag, $pre, \%options)
 sub _match_tagged	# ($$$$$$$)
 {
 	my ($textref, $pre, $ldel, $rdel, $omode, $bad, $ignore) = @_;
+	my $rdelspec;
 
 	my ($startpos, $opentagpos, $textpos, $parapos, $closetagpos, $endpos) = ( pos($$textref) = pos($$textref)||0 );
 
@@ -290,7 +291,7 @@ sub _match_tagged	# ($$$$$$$)
 
 	$opentagpos = pos($$textref);
 
-	unless ($$textref =~ m/\G($ldel)/gc)
+	unless ($$textref =~ m/\G$ldel/gc)
 	{
 		$@ = "Did not find opening tag: /$ldel/";
 		goto failed;
@@ -300,13 +301,17 @@ sub _match_tagged	# ($$$$$$$)
 
 	if (!defined $rdel)
 	{
-		$rdel = $1;
-		unless ($rdel =~ s/\A([[(<{]+)($XMLNAME).*/ "$1\/$2". revbracket($1) /oes)
+		$rdelspec = $&;
+		unless ($rdelspec =~ s/\A([[(<{]+)($XMLNAME).*/ quotemeta "$1\/$2". revbracket($1) /oes)
 		{
 			$@ = "Unable to construct closing tag to match: $rdel";
 			goto failed;
 		}
 	}
+	else
+	{
+		$rdelspec = eval "qq{$rdel}";
+	}
 
 	while (pos($$textref) < length($$textref))
 	{
@@ -317,7 +322,7 @@ sub _match_tagged	# ($$$$$$$)
 			$parapos = pos($$textref) - length($1)
 				unless defined $parapos;
 		}
-		elsif ($$textref =~ m/\G($rdel)/gc )
+		elsif ($$textref =~ m/\G($rdelspec)/gc )
 		{
 			$closetagpos = pos($$textref)-length($1);
 			goto matched;
@@ -403,7 +408,7 @@ sub _match_variable($$)
 		return;
 	}
 	my $varpos = pos($$textref);
-	unless ($$textref =~ m/\G(\$#?|[\@\%])+/gc)
+	unless ($$textref =~ m/\G(\$#?|[*\@\%]|\\&)+/gc)
 	{
 		$@ = "Did not find leading dereferencer";
 		pos $$textref = $startpos;
@@ -418,14 +423,19 @@ sub _match_variable($$)
 		return;
 	}
 
-	1 while ( _match_codeblock($textref,'\s*(?:->(?:\s*\w+\s*)?)?\s*',
-				   '[({[]','[)}\]]',
-				   '[({[]','[)}\]]',
-				   0
-				   )
-	        || _match_variable($textref,'\s*->\s*')
-	        || $$textref =~ m/\G\s*->\s*\w+(?![{([])/gc
-	        );
+	while (1)
+	{
+		next if _match_codeblock($textref,
+					 qr/\s*->\s*(?:[a-zA-Z]\w+\s*)?/,
+					 qr/[({[]/, qr/[)}\]]/,
+					 qr/[({[]/, qr/[)}\]]/, 0);
+		next if _match_codeblock($textref,
+					 qr/\s*/, qr/[{[]/, qr/[}\]]/,
+					 qr/[{[]/, qr/[}\]]/, 0);
+		next if _match_variable($textref,'\s*->\s*');
+		next if $$textref =~ m/\G\s*->\s*\w+(?![{([])/gc;
+		last;
+	}
 	
 	my $endpos = pos($$textref);
 	return ($startpos, $varpos-$startpos,
@@ -514,17 +524,23 @@ sub _match_codeblock($$$$$$$)
 			last;
 		}
 
-		# NEED TO COVER MANY MORE CASES HERE!!!
-		if ($$textref =~ m#\G\s*(=~|!~|split|grep|map|return|;|[|]{1,2}|[&]{1,2})#gc)
+		if (_match_variable($textref,'\s*') ||
+		    _match_quotelike($textref,'\s*',$patvalid,$patvalid) )
 		{
-			$patvalid = 1;
+			$patvalid = 0;
 			next;
 		}
 
-		if (_match_variable($textref,'\s*') ||
-		    _match_quotelike($textref,'\s*',$patvalid) )
+
+		# NEED TO COVER MANY MORE CASES HERE!!!
+		if ($$textref =~ m#\G\s*( [-+*x/%^&|.]=?
+					| =(?!>)
+					| (\*\*|&&|\|\||<<|>>)=?
+					| [!=][~=]
+					| split|grep|map|return
+					)#gcx)
 		{
-			$patvalid = 0;
+			$patvalid = 1;
 			next;
 		}
 
@@ -543,7 +559,7 @@ sub _match_codeblock($$$$$$$)
 		}
 
 		$patvalid = 0;
-		$$textref =~ m/\G\s*(\w+|[-=>]>|.)/gc;
+		$$textref =~ m/\G\s*(\w+|[-=>]>|.|\Z)/gc;
 	}
 
 	unless ($matched)
@@ -579,7 +595,7 @@ sub extract_quotelike (;$$)
 	my $wantarray = wantarray;
 	my $pre  = defined $_[1] ? $_[1] : '\s*';
 
-	my @match = _match_quotelike($textref,$pre,0);
+	my @match = _match_quotelike($textref,$pre,1,0);
 	return _fail($wantarray, $textref) unless @match;
 	return _succeed($wantarray, $textref,
 			$match[2], $match[18]-$match[2],	# MATCH
@@ -589,9 +605,9 @@ sub extract_quotelike (;$$)
 		       );
 };
 
-sub _match_quotelike($$$)	# ($textref, $prepat, $allow_qmark)
+sub _match_quotelike($$$$)	# ($textref, $prepat, $allow_raw_match)
 {
-	my ($textref, $pre, $qmark) = @_;
+	my ($textref, $pre, $rawmatch, $qmark) = @_;
 
 	my ($textlen,$startpos,
 	    $oppos,
@@ -610,7 +626,9 @@ sub _match_quotelike($$$)	# ($textref, $prepat, $allow_qmark)
 
 	my $initial = substr($$textref,$oppos,1);
 
-	if ($initial && $initial =~ m|^[/\"\'\`]| || $qmark && $initial =~ m|^\?|)
+	if ($initial && $initial =~ m|^[\"\'\`]|
+		     || $rawmatch && $initial =~ m|^/|
+		     || $qmark && $initial =~ m|^\?|)
 	{
 		unless ($$textref =~ m/ \Q$initial\E [^\\$initial]* (\\.[^\\$initial]*)* \Q$initial\E /gcx)
 		{
@@ -679,7 +697,8 @@ sub _match_quotelike($$$)	# ($textref, $prepat, $allow_qmark)
 	}
 	$ld2pos = $rd1pos = pos($$textref)-1;
 
-	if ($op =~ /s|tr|y/)
+	my $second_arg = $op =~ /s|tr|y/ ? 1 : 0;
+	if ($second_arg)
 	{
 		my ($ldel2, $rdel2);
 		if ($ldel1 =~ /[[(<{]/)
@@ -728,9 +747,9 @@ sub _match_quotelike($$$)	# ($textref, $prepat, $allow_qmark)
 		$ld1pos,	1,			# LEFT DEL
 		$str1pos,	$rd1pos-$str1pos,	# STR/PAT
 		$rd1pos,	1,			# RIGHT DEL
-		$ld2pos,	1,			# 2ND LDEL
-		$str2pos,	$rd2pos-$str2pos,	# 2ND STR
-		$rd2pos,	1,			# 2ND RDEL
+		$ld2pos,	$second_arg,		# 2ND LDEL (MAYBE)
+		$str2pos,	$rd2pos-$str2pos,	# 2ND STR (MAYBE)
+		$rd2pos,	$second_arg,		# 2ND RDEL (MAYBE)
 		$modpos,	$endpos-$modpos,	# MODIFIERS
 		$endpos,	$textlen-$endpos,	# REMAINDER
 	       );
@@ -768,16 +787,34 @@ sub extract_multiple (;$$$$)	# ($text, $functions_ref, $max_fields, $ignoreunkno
 
 		my $unkpos;
 		my $func;
-		my $field;
+		my $class;
+
+		my @class;
+		foreach $func ( @func )
+		{
+			if (ref($func) eq 'HASH')
+			{
+				push @class, (keys %$func)[0];
+				$func = (values %$func)[0];
+			}
+			else
+			{
+				push @class, undef;
+			}
+		}
 
 		FIELD: while (pos() < length())
 		{
-			foreach $func ( @func )
+			my $field;
+			foreach my $i ( 0..$#func )
 			{
-				undef $field;
+				$func = $func[$i];
+				$class = $class[$i];
 				$lastpos = pos;
 				if (ref($func) eq 'CODE')
 					{ ($field) = $func->($_) }
+				elsif (ref($func) eq 'Text::Balanced::Extractor')
+					{ $field = $func->extract($_) }
 				elsif( m/\G$func/gc )
 					{ $field = defined($1) ? $1 : $& }
 
@@ -790,7 +827,9 @@ sub extract_multiple (;$$$$)	# ($text, $functions_ref, $max_fields, $ignoreunkno
 						undef $unkpos;
 						last FIELD if @fields == $max;
 					}
-					push @fields, $field;
+					push @fields, $class 
+						? bless(\$field, $class)
+						: $field;
 					$firstpos = $lastpos unless defined $firstpos;
 					$lastpos = pos;
 					last FIELD if @fields == $max;
